from __future__ import annotations

import inspect
from typing import Any, Dict, Iterable, List, Optional, overload

from di._docstrings import join_docstring_from
from di._inspect import get_parameters, infer_call_from_annotation
from di.exceptions import WiringError
from di.types.dependencies import DependantProtocol, DependencyParameter
from di.types.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProviderType,
    DependencyType,
    GeneratorProvider,
)
from di.types.scopes import Scope

_VARIABLE_PARAMETER_KINDS = (
    inspect.Parameter.VAR_POSITIONAL,
    inspect.Parameter.VAR_KEYWORD,
)


class Dependant(DependantProtocol[DependencyType]):
    wire: bool
    autowire: bool

    @overload
    def __init__(
        self,
        call: Optional[AsyncGeneratorProvider[DependencyType]] = None,
        *,
        scope: Optional[Scope] = None,
        share: bool = True,
        wire: bool = True,
        autowire: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CoroutineProvider[DependencyType]] = None,
        *,
        scope: Optional[Scope] = None,
        share: bool = True,
        wire: bool = True,
        autowire: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[GeneratorProvider[DependencyType]] = None,
        *,
        scope: Optional[Scope] = None,
        share: bool = True,
        wire: bool = True,
        autowire: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CallableProvider[DependencyType]] = None,
        *,
        scope: Optional[Scope] = None,
        share: bool = True,
        wire: bool = True,
        autowire: bool = True,
    ) -> None:
        ...

    def __init__(
        self,
        call: Optional[DependencyProviderType[DependencyType]] = None,
        *,
        scope: Scope = None,
        share: bool = True,
        wire: bool = True,
        autowire: bool = True,
    ) -> None:
        self.call = call
        self.scope = scope
        self.dependencies: Optional[
            List[DependencyParameter[DependantProtocol[Any]]]
        ] = None
        self.share = share
        self.autowire = autowire
        self.wire = wire

    def __repr__(self) -> str:
        share = "" if self.share is False else ", share=True"
        return f"{self.__class__.__name__}(call={self.call}, scope={self.scope}{share})"

    @join_docstring_from(DependantProtocol[Any].__hash__)
    def __hash__(self) -> int:
        return id(self.call)

    @join_docstring_from(DependantProtocol[Any].__eq__)
    def __eq__(self, o: object) -> bool:
        if type(self) is not type(o):
            return False
        assert isinstance(o, type(self))
        if self.share is False or o.share is False:
            return False
        return self.call is o.call

    @join_docstring_from(DependantProtocol[Any].get_dependencies)
    def get_dependencies(
        self,
    ) -> List[DependencyParameter[DependantProtocol[Any]]]:
        """For the Dependant implementation, this serves as a cache layer on
        top of gather_dependencies.
        """
        if self.dependencies is None:
            self.dependencies = self.gather_dependencies()
        return self.dependencies

    def gather_parameters(self) -> Dict[str, inspect.Parameter]:
        """Collect parameters that this dependency needs to construct itself.

        Generally, this means introspecting into our own callable (self.call).
        """
        assert self.call is not None, "Cannot gather parameters without a bound call"
        return get_parameters(self.call)

    def register_parameter(self, param: inspect.Parameter) -> None:
        if self.call is None:
            self.call = infer_call_from_annotation(param)

    def gather_dependencies(
        self,
    ) -> List[DependencyParameter[DependantProtocol[Any]]]:
        """Collect this dependencies sub dependencies.

        The returned dict corresponds to keyword arguments that will be passed
        to this dependencies `call` after all sub-dependencies are themselves resolved.
        """
        if self.wire is False:
            return []
        assert (
            self.call is not None
        ), "Container should have assigned call; this is a bug!"
        res: List[DependencyParameter[DependantProtocol[Any]]] = []
        for param in self.gather_parameters().values():
            if param.kind in _VARIABLE_PARAMETER_KINDS:
                raise WiringError(
                    "Dependencies may not use variable positional or keyword arguments"
                )
            if isinstance(param.default, Dependant):
                sub_dependant: DependantProtocol[Any] = param.default
            elif param.default is param.empty and self.autowire:
                sub_dependant = self.create_sub_dependant(param)
            else:
                continue  # pragma: no cover
            sub_dependant.register_parameter(param)
            res.append(DependencyParameter(dependency=sub_dependant, parameter=param))
        return res

    def create_sub_dependant(self, param: inspect.Parameter) -> DependantProtocol[Any]:
        """Create a Dependant instance from a sub-dependency of this Dependency.

        This is used in the scenario where a transient dependency is inferred from a type hint.
        For example:

        >>> class Foo:
        >>>     ...
        >>> def foo_factory(foo: Foo) -> Foo:
        >>>     return foo
        >>> def parent(foo: Dependency(foo_factory)):
        >>>    ...

        In this scenario, `Dependency(foo_factory)` will call `create_sub_dependant(Foo)`.

        It is recommended to transfer `scope` and possibly `share` to sub-dependencies created in this manner.
        """
        if param.annotation is inspect.Parameter.empty:
            raise WiringError(
                "Cannot wire a parameter with no default and no type annotation"
            )
        return Dependant[Any](
            call=param.annotation,
            scope=self.scope,
            share=self.share,
            autowire=self.autowire,
        )


class JoinedDependant(Dependant[DependencyType]):
    def __init__(
        self,
        dependant: DependantProtocol[DependencyType],
        *,
        siblings: Iterable[DependantProtocol[Any]],
        scope: Scope = None,
        share: bool = True,
        autowire: bool = True,
    ) -> None:
        self.call = dependant.call
        self.dependant = dependant
        self.siblings = siblings
        self.scope = scope
        self.dependencies: Optional[
            List[DependencyParameter[DependantProtocol[Any]]]
        ] = None
        self.share = share
        self.autowire = autowire

    def gather_dependencies(self) -> List[DependencyParameter[DependantProtocol[Any]]]:
        return [
            *self.dependant.get_dependencies(),
            *(DependencyParameter(dep, None) for dep in self.siblings),
        ]



from __future__ import annotations

import functools
from collections import defaultdict, deque
from contextvars import ContextVar
from typing import (
    Any,
    ContextManager,
    DefaultDict,
    Deque,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
    cast,
)

from di._dag import topsort
from di._inspect import is_async_gen_callable, is_coroutine_callable
from di._local_scope_context import LocalScopeContext
from di._state import ContainerState
from di._task import AsyncTask, SyncTask, Task
from di.exceptions import (
    DependencyRegistryError,
    DuplicateScopeError,
    ScopeViolationError,
    UnknownScopeError,
)
from di.executors import DefaultExecutor
from di.types import FusedContextManager
from di.types.dependencies import DependantProtocol, DependencyParameter
from di.types.executor import AsyncExecutor, SyncExecutor
from di.types.executor import Task as ExecutorTask
from di.types.providers import (
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
)
from di.types.scopes import Scope
from di.types.solved import SolvedDependency

Dependency = Any


class Container:
    _context: ContextVar[ContainerState]
    _executor: Union[AsyncExecutor, SyncExecutor]

    def __init__(
        self,
        *,
        default_scope: Scope = None,
        executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None,
    ) -> None:
        self._context = ContextVar("context")
        state = ContainerState()
        state.cached_values.add_scope("container")
        state.cached_values.set(Container, self, scope="container")
        self._context.set(state)
        self._executor = executor or DefaultExecutor()
        self._default_scope = default_scope

    @property
    def _state(self) -> ContainerState:
        return self._context.get()

    @property
    def scopes(self) -> List[Scope]:
        return self._state.scopes

    def enter_global_scope(self, scope: Scope) -> FusedContextManager[None]:
        """Enter a global scope that is share amongst threads and coroutines.

        If you enter a global scope in one thread / coroutine, it will propagate to others.
        """
        return self._state.enter_scope(scope)

    def enter_local_scope(self, scope: Scope) -> FusedContextManager[None]:
        """Enter a local scope that is localized to the current thread or coroutine.

        If you enter a global scope in one thread / coroutine, it will NOT propagate to others.
        """
        if scope in self._state.stacks:
            raise DuplicateScopeError(f"Scope {scope} has already been entered!")
        return LocalScopeContext(self._context, scope)

    def bind(
        self,
        provider: DependantProtocol[DependencyType],
        dependency: DependencyProviderType[DependencyType],
    ) -> ContextManager[None]:
        """Bind a new dependency provider for a given dependency.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).

        Binds are only identified by the identity of the callable and do not take into account
        the scope or any other data from the dependency they are replacing.

        The `scope` parameter determines the scope for the bind itself.
        The bind will be automatically cleared when that scope is exited.
        If no scope is provided, the current scope is used.
        """
        return self._state.bind(provider=provider, dependency=dependency)

    def __contains__(self, o: object) -> bool:
        return o in self._state.binds

    def __getitem__(
        self, provider: DependencyProviderType[DependencyType]
    ) -> DependantProtocol[DependencyType]:
        return self._state.binds[provider]

    def solve(
        self,
        dependency: DependantProtocol[DependencyType],
    ) -> SolvedDependency[DependencyType]:
        """Solve a dependency.

        Returns a SolvedDependency that can be executed to get the dependency's value.

        The `siblings` paramter can be used to pass additional dependencies that should be executed
        that aren't strictly needed to solve `dependency`.
        """

        if dependency.call in self._state.binds:
            dependency = self._state.binds[dependency.call]

        param_graph: Dict[
            DependantProtocol[Any],
            List[DependencyParameter[DependantProtocol[Any]]],
        ] = {}

        dep_registry: Dict[DependantProtocol[Any], DependantProtocol[Any]] = {}

        dep_dag: Dict[DependantProtocol[Any], List[DependantProtocol[Any]]] = {}

        def get_params(
            dep: DependantProtocol[Any],
        ) -> List[DependencyParameter[DependantProtocol[Any]]]:
            params = dep.get_dependencies().copy()
            for idx, param in enumerate(params):
                assert param.dependency.call is not None
                if param.dependency.call in self._state.binds:
                    params[idx] = DependencyParameter[Any](
                        dependency=self._state.binds[param.dependency.call],
                        parameter=param.parameter,
                    )
            return params

        def check_equivalent(dep: DependantProtocol[Any]):
            if dep in dep_registry and dep.scope != dep_registry[dep].scope:
                raise DependencyRegistryError(
                    f"The dependencies {dep} and {dep_registry[dep]}"
                    " have the same lookup (__hash__ and __eq__) but have different scopes"
                    f" ({dep.scope} and {dep_registry[dep].scope} respectively)"
                    " This is likely a mistake, but you can override this behavior by either:"
                    "\n  1. Wrapping the function in another function or subclassing the class such"
                    " that they now are considered different dependencies"
                    "\n  2. Use a custom implemetnation of DependantProtocol that changes the meaning"
                    " or computation of __hash__ and/or __eq__"
                )

        q: Deque[DependantProtocol[Any]] = deque([dependency])
        while q:
            dep = q.popleft()
            if dep in dep_registry:
                check_equivalent(dep)
            else:
                dep_registry[dep] = dep
                params = get_params(dep)
                param_graph[dep] = params
                dep_dag[dep] = []
                for param in params:
                    subdep = param.dependency
                    dep_dag[dep].append(subdep)
                    if subdep not in dep_registry:
                        q.append(subdep)
        tasks = self._build_tasks(param_graph, topsort(dep_dag))
        dependency_dag = {dep: set(subdeps) for dep, subdeps in dep_dag.items()}
        # we store several things:
        # - the dependency itself
        # - dag, which includes parameter information and is public
        # - _dependency_dag: same as dag but without param info and w/o repeated params (so a set instead of a list)
        # - _tasks: a mapping of dependency to tasks
        return SolvedDependency(
            dependency=dependency,
            dag=param_graph,
            _dependency_dag=dependency_dag,
            _tasks=tasks,
        )

    def _build_tasks(
        self,
        dag: Dict[
            DependantProtocol[Any],
            List[DependencyParameter[DependantProtocol[Any]]],
        ],
        topsorted: List[DependantProtocol[Any]],
    ) -> Dict[DependantProtocol[Any], Union[AsyncTask[Any], SyncTask[Any]]]:
        tasks: Dict[DependantProtocol[Any], Union[AsyncTask[Any], SyncTask[Any]]] = {}
        for dep in reversed(topsorted):
            tasks[dep] = self._build_task(dep, tasks, dag)
        return tasks

    def _build_task(
        self,
        dependency: DependantProtocol[Any],
        tasks: Dict[DependantProtocol[Any], Union[AsyncTask[Any], SyncTask[Any]]],
        dag: Dict[
            DependantProtocol[Any],
            List[DependencyParameter[DependantProtocol[Any]]],
        ],
    ) -> Union[AsyncTask[Any], SyncTask[Any]]:

        task_dependencies: List[DependencyParameter[Task[Any]]] = [
            DependencyParameter(
                dependency=tasks[param.dependency], parameter=param.parameter
            )
            for param in dag[dependency]
        ]

        if is_async_gen_callable(dependency.call) or is_coroutine_callable(
            dependency.call
        ):
            return AsyncTask(dependant=dependency, dependencies=task_dependencies)
        else:
            return SyncTask(dependant=dependency, dependencies=task_dependencies)

    def _validate_scopes(self, solved: SolvedDependency[Dependency]) -> None:
        """Validate that dependencies all have a valid scope and
        that dependencies only depend on outer scopes or their own scope.
        """
        scopes: Dict[Scope, int] = {
            scope: idx
            for idx, scope in enumerate(reversed(self._state.scopes + [None]))
        }

        def check_is_inner(
            dep: DependantProtocol[Any], subdep: DependantProtocol[Any]
        ) -> None:
            if scopes[dep.scope] > scopes[subdep.scope]:
                raise ScopeViolationError(
                    f"{dep} cannot depend on {subdep} because {subdep}'s"
                    f" scope ({subdep.scope}) is narrower than {dep}'s scope ({dep.scope})"
                )

        def check_scope(dep: DependantProtocol[Any]) -> None:
            if dep.scope not in scopes:
                raise UnknownScopeError(
                    f"Dependency{dep} has an unknown scope {dep.scope}."
                    f" Did you forget to enter the {dep.scope} scope?"
                )

        for dep, params in solved.dag.items():
            check_scope(dep)
            for param in params:
                subdep = param.dependency
                check_scope(subdep)
                check_is_inner(dep, subdep)

    def _prepare_execution(
        self,
        solved: SolvedDependency[Any],
        *,
        validate_scopes: bool = True,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> Tuple[Dict[DependantProtocol[Any], Any], List[ExecutorTask]]:
        results: Dict[DependantProtocol[Any], Any] = {}
        values = values or {}
        if validate_scopes:
            self._validate_scopes(solved)

        tasks = solved._tasks

        # Build a DAG of Tasks that we actually need to execute
        # this allows us to prune subtrees that will come from cached values
        unvisited = deque([solved.dependency])
        # Make DAG values a set to account for dependencies that depend on the the same
        # sub dependency in more than one param (`def func(a: A, a_again: A)`)
        dag: Dict[DependantProtocol[Any], List[DependantProtocol[Any]]] = {}
        dependant_dag: DefaultDict[
            DependantProtocol[Any], List[Task[Any]]
        ] = defaultdict(list)
        while unvisited:
            dep = unvisited.pop()
            if dep in dag:
                continue
            task = tasks[dep]
            # task the dependency is cached or was provided by value
            # we don't need to compute it or any of it's dependencies
            if not task.from_cache_or_values(self._state, results, values):
                # otherwise, we add it to our DAG and visit it's children
                dag[dep] = []
                for subdep in solved._dependency_dag[dep]:
                    if not tasks[subdep].from_cache_or_values(
                        self._state, results, values
                    ):
                        dependant_dag[subdep].append(tasks[dep])
                        dag[dep].append(subdep)
                        if subdep not in dag:
                            unvisited.append(subdep)

        dependency_counts = {dep: len(subdeps) for dep, subdeps in dag.items()}

        leafs: List[ExecutorTask] = []
        for dep, depcount in dependency_counts.items():
            if depcount == 0:
                leaf = functools.partial(
                    tasks[dep].compute,
                    state=self._state,
                    results=results,
                    values=values,
                    dependency_counts=dependency_counts,
                    dependants=dependant_dag,
                )
                leafs.append(leaf)  # type: ignore[arg-type] # beacuse of partial

        return results, leafs

    def execute_sync(
        self,
        solved: SolvedDependency[DependencyType],
        *,
        validate_scopes: bool = True,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        with self.enter_local_scope(self._default_scope):
            results, queue = self._prepare_execution(
                solved, validate_scopes=validate_scopes, values=values
            )
            if not hasattr(self._executor, "execute_sync"):
                raise TypeError(
                    "execute_sync requires an executor implementing the SyncExecutor protocol"
                )
            executor = cast(SyncExecutor, self._executor)
            if queue:
                executor.execute_sync(queue)
            return results[solved.dependency]

    async def execute_async(
        self,
        solved: SolvedDependency[DependencyType],
        *,
        validate_scopes: bool = True,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        async with self.enter_local_scope(self._default_scope):
            results, queue = self._prepare_execution(
                solved, validate_scopes=validate_scopes, values=values
            )
            if not hasattr(self._executor, "execute_async"):
                raise TypeError(
                    "execute_async requires an executor implementing the AsyncExecutor protocol"
                )
            executor = cast(AsyncExecutor, self._executor)
            if queue:
                await executor.execute_async(queue)
            return results[solved.dependency]



import sys
from types import TracebackType
from typing import Optional, Type, TypeVar, Union

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

T = TypeVar("T", covariant=True)


class FusedContextManager(Protocol[T]):
    __slots__ = ()

    def __enter__(self):
        ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Union[None, bool]:
        ...

    async def __aenter__(self):
        ...

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Union[None, bool]:
        ...



from __future__ import annotations

import sys
from typing import Awaitable, Iterable, Optional, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol


class Task(Protocol):
    def __call__(
        self,
    ) -> Union[Awaitable[Iterable[Optional[Task]]], Iterable[Optional[Task]]]:
        ...


class SyncExecutor(Protocol):
    def execute_sync(self, tasks: Iterable[Task]) -> None:
        raise NotImplementedError


class AsyncExecutor(Protocol):
    async def execute_async(self, tasks: Iterable[Task]) -> None:
        raise NotImplementedError



from __future__ import annotations

import functools
import typing
from collections import deque
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Deque,
    Dict,
    Generic,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Union,
    cast,
)

from di._inspect import is_async_gen_callable, is_gen_callable
from di._state import ContainerState
from di.exceptions import IncompatibleDependencyError
from di.types.dependencies import DependantProtocol, DependencyParameter
from di.types.executor import Task as ExecutorTask
from di.types.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProvider,
    DependencyType,
    GeneratorProvider,
)


class Task(Generic[DependencyType]):
    __slots__ = ("dependant", "dependencies")

    def __init__(
        self,
        dependant: DependantProtocol[DependencyType],
        dependencies: List[DependencyParameter[Task[Any]]],
    ) -> None:
        self.dependant = dependant
        self.dependencies = dependencies

    def compute(
        self,
        state: ContainerState,
        results: Dict[DependantProtocol[Any], Any],
        values: typing.Mapping[DependencyProvider, Any],
        dependency_counts: MutableMapping[DependantProtocol[Any], int],
        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
    ) -> Union[
        Awaitable[Iterable[Optional[ExecutorTask]]], Iterable[Optional[ExecutorTask]]
    ]:
        raise NotImplementedError

    def from_cache_or_values(
        self,
        state: ContainerState,
        results: Dict[DependantProtocol[Any], Any],
        values: typing.Mapping[DependencyProvider, Any],
    ) -> bool:
        assert self.dependant.call is not None
        if self.dependant.call in values:
            value = values[self.dependant.call]
            results[self.dependant] = value
            if self.dependant.share:
                state.cached_values.set(
                    self.dependant.call, value, scope=self.dependant.scope
                )
            return True
        elif self.dependant.share and state.cached_values.contains(self.dependant.call):
            results[self.dependant] = state.cached_values.get(self.dependant.call)
            return True
        return False

    def gather_params(
        self, results: Dict[DependantProtocol[Any], Any]
    ) -> Tuple[List[Any], Dict[str, Any]]:
        positional: List[Any] = []
        keyword: Dict[str, Any] = {}
        for dep in self.dependencies:
            if dep.parameter is not None:
                if dep.parameter.kind is dep.parameter.kind.POSITIONAL_ONLY:
                    positional.append(results[dep.dependency.dependant])
                else:
                    keyword[dep.parameter.name] = results[dep.dependency.dependant]
        return positional, keyword

    def gather_new_tasks(
        self,
        state: ContainerState,
        results: Dict[DependantProtocol[Any], Any],
        values: typing.Mapping[DependencyProvider, Any],
        dependency_counts: MutableMapping[DependantProtocol[Any], int],
        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
    ) -> Iterable[Optional[ExecutorTask]]:
        """Look amongst our dependants to see if any of them are now dependency free"""
        new_tasks: Deque[Optional[ExecutorTask]] = deque()
        for dependant in dependants[self.dependant]:
            count = dependency_counts[dependant.dependant]
            if count == 1:
                # this dependant has no further dependencies, so we can compute it now
                newtask = functools.partial(
                    dependant.compute,
                    state=state,
                    results=results,
                    values=values,
                    dependency_counts=dependency_counts,
                    dependants=dependants,
                )
                new_tasks.append(newtask)  # type: ignore[arg-type] # because of partial
                # pop it from dependency counts so that we can
                # tell when we've satisfied all dependencies below
                dependency_counts.pop(dependant.dependant)
            else:
                dependency_counts[dependant.dependant] = count - 1
        # also pop ourselves if we have no deps
        if dependency_counts.get(self.dependant, -1) == 0:
            dependency_counts.pop(self.dependant)
        if not dependency_counts:
            # all dependencies are taken care of, insert sentinel None value
            new_tasks.append(None)
        return new_tasks

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.dependant)}, {repr(self.dependencies)})"


class AsyncTask(Task[DependencyType]):
    __slots__ = ()

    async def compute(
        self,
        state: ContainerState,
        results: Dict[DependantProtocol[Any], Any],
        values: typing.Mapping[DependencyProvider, Any],
        dependency_counts: MutableMapping[DependantProtocol[Any], int],
        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
    ) -> Iterable[Optional[ExecutorTask]]:
        args, kwargs = self.gather_params(results)

        assert self.dependant.call is not None
        call = self.dependant.call
        if is_async_gen_callable(call):
            stack = state.stacks[self.dependant.scope]
            if not isinstance(stack, AsyncExitStack):
                raise IncompatibleDependencyError(
                    f"The dependency {self.dependant} is an awaitable dependency"
                    f" and canot be used in the sync scope {self.dependant.scope}"
                )
            if TYPE_CHECKING:
                call = cast(AsyncGeneratorProvider[DependencyType], call)
            results[self.dependant] = await stack.enter_async_context(
                asynccontextmanager(call)(*args, **kwargs)
            )
        else:
            if TYPE_CHECKING:
                call = cast(CoroutineProvider[DependencyType], call)
            results[self.dependant] = await call(*args, **kwargs)
        if self.dependant.share:
            # caching is allowed, now that we have a value we can save it and start using the cache
            state.cached_values.set(
                self.dependant.call, results[self.dependant], scope=self.dependant.scope
            )

        return self.gather_new_tasks(
            state=state,
            results=results,
            values=values,
            dependency_counts=dependency_counts,
            dependants=dependants,
        )


class SyncTask(Task[DependencyType]):
    __slots__ = ()

    def compute(
        self,
        state: ContainerState,
        results: Dict[DependantProtocol[Any], Any],
        values: typing.Mapping[DependencyProvider, Any],
        dependency_counts: MutableMapping[DependantProtocol[Any], int],
        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
    ) -> Iterable[Optional[ExecutorTask]]:

        args, kwargs = self.gather_params(results)

        assert self.dependant.call is not None
        call = self.dependant.call
        if is_gen_callable(self.dependant.call):
            if TYPE_CHECKING:
                call = cast(GeneratorProvider[DependencyType], call)
            stack = state.stacks[self.dependant.scope]
            results[self.dependant] = stack.enter_context(
                contextmanager(call)(*args, **kwargs)
            )
        else:
            if TYPE_CHECKING:
                call = cast(CallableProvider[DependencyType], call)
            results[self.dependant] = call(*args, **kwargs)
        if self.dependant.share:
            # caching is allowed, now that we have a value we can save it and start using the cache
            state.cached_values.set(
                self.dependant.call, results[self.dependant], scope=self.dependant.scope
            )

        return self.gather_new_tasks(
            state=state,
            results=results,
            values=values,
            dependency_counts=dependency_counts,
            dependants=dependants,
        )



try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    # python <3.8
    import importlib_metadata  # type: ignore

try:
    __version__ = importlib_metadata.version(__name__)  # type: ignore # for py 3.7
except importlib_metadata.PackageNotFoundError:  # type: ignore # for py 3.7
    __version__ = "0.0.0"


from di.container import Container
from di.dependant import Dependant
from di.params import Depends

__all__ = ("Container", "Dependant", "Depends")



from __future__ import annotations

import dataclasses
from typing import Any, Generic, List, Mapping, Set

from di._task import Task
from di.types.dependencies import DependantProtocol, DependencyParameter
from di.types.providers import DependencyType

Dependency = Any


@dataclasses.dataclass
class SolvedDependency(Generic[DependencyType]):
    """Representation of a fully solved dependency as DAG"""

    dependency: DependantProtocol[DependencyType]
    dag: Mapping[
        DependantProtocol[Any], List[DependencyParameter[DependantProtocol[Any]]]
    ]
    # the following attributes are used to cache data for solved dependencies
    # they are put on SolvedDependency so that they can be garbage collected along with it
    # these are implementation details and are subject to change at any time
    _tasks: Mapping[DependantProtocol[Any], Task[Any]]
    _dependency_dag: Mapping[DependantProtocol[Any], Set[DependantProtocol[Any]]]

    def get_flat_subdependants(self) -> List[DependantProtocol[Any]]:
        """Get an exhaustive list of all of the dependencies of this dependency,
        in no particular order.
        """
        return list(self.dag.keys() - {self.dependency})



import concurrent.futures
import inspect
import typing
from collections import deque

import anyio
import anyio.abc

from di._concurrency import curry_context, guarantee_awaitable
from di._inspect import is_coroutine_callable
from di.types.executor import AsyncExecutor, SyncExecutor, Task


def _check_not_coro(
    task: typing.Optional[Task],
) -> typing.Callable[[], typing.Iterable[typing.Optional[Task]]]:
    if task is not None and is_coroutine_callable(task):
        raise TypeError("Cannot execute async dependencies in execute_sync")
    return task  # type: ignore[return-value]


_AsyncTaskRetval = typing.Awaitable[typing.Iterable[Task]]
_SyncTaskRetval = typing.Iterable[Task]


class SimpleSyncExecutor(SyncExecutor):
    def execute_sync(self, tasks: typing.Iterable[Task]) -> None:
        q: typing.Deque[typing.Optional[Task]] = deque(tasks)
        while q:
            task = q.popleft()
            if task is None:
                return
            newtasks = _check_not_coro(task)()
            assert not isinstance(newtasks, typing.Awaitable)
            q.extend(newtasks)


class SimpleAsyncExecutor(AsyncExecutor):
    async def execute_async(self, tasks: typing.Iterable[Task]) -> None:
        q: typing.Deque[typing.Optional[Task]] = deque(tasks)
        while q:
            task = q.popleft()
            if task is None:
                return
            maybe_coro = task()
            if inspect.isawaitable(maybe_coro):
                newtasks = await typing.cast(_AsyncTaskRetval, maybe_coro)
            else:
                newtasks = typing.cast(_SyncTaskRetval, maybe_coro)
            q.extend(newtasks)


class ConcurrentSyncExecutor(AsyncExecutor):
    def execute_sync(self, tasks: typing.Iterable[Task]) -> None:
        futures: typing.Set[
            concurrent.futures.Future[typing.Iterable[typing.Optional[Task]]]
        ] = set()
        with concurrent.futures.ThreadPoolExecutor() as exec:
            for task in tasks:
                futures.add(exec.submit(curry_context(_check_not_coro(task))))
            while futures:
                for future in concurrent.futures.as_completed(futures):
                    newtasks = future.result()
                    futures.remove(future)
                    if inspect.isawaitable(newtasks):
                        raise TypeError(
                            "Cannot execute async dependencies in execute_sync"
                        )
                    for newtask in newtasks:
                        if newtask is None:
                            break
                        futures.add(
                            exec.submit(curry_context(_check_not_coro(newtask)))
                        )


async def _async_worker(
    task: Task,
    stream: anyio.abc.ObjectSendStream[typing.Optional[Task]],
) -> None:
    try:
        newtasks = typing.cast(_SyncTaskRetval, await guarantee_awaitable(task)())
    except Exception:
        try:
            await stream.send(None)
        except anyio.ClosedResourceError:
            pass
        raise
    for newtask in newtasks:
        try:
            await stream.send(newtask)
        except anyio.ClosedResourceError:
            pass


Streams = typing.Tuple[
    anyio.abc.ObjectSendStream[typing.Optional[Task]],
    anyio.abc.ObjectReceiveStream[typing.Optional[Task]],
]


class ConcurrentAsyncExecutor(AsyncExecutor):
    async def execute_async(self, tasks: typing.Iterable[Task]) -> None:
        streams = typing.cast(Streams, anyio.create_memory_object_stream(float("inf")))
        send, receive = streams
        for task in tasks:
            await send.send(task)
        async with anyio.create_task_group() as taskgroup, send, receive:
            while True:
                newtask = await receive.receive()
                if newtask is None:
                    return None
                taskgroup.start_soon(_async_worker, newtask, send)


class DefaultExecutor(ConcurrentSyncExecutor, ConcurrentAsyncExecutor):
    ...



import functools
import inspect
from functools import lru_cache, wraps
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Coroutine,
    Dict,
    Generator,
    Mapping,
    Optional,
    TypeVar,
    Union,
    get_type_hints,
)

from di.exceptions import WiringError

CallableProvider = Callable[..., Any]
CoroutineProvider = Callable[..., Coroutine[Any, Any, Any]]
GeneratorProvider = Callable[..., Generator[Any, None, None]]
AsyncGeneratorProvider = Callable[..., AsyncGenerator[Any, None]]

DependencyProvider = Union[
    AsyncGeneratorProvider,
    CoroutineProvider,
    GeneratorProvider,
    CallableProvider,
]


T = TypeVar("T")


def cached_accept_callable_class(
    maxsize: int,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def wrapper(func: Callable[..., T]) -> Callable[..., T]:
        func = lru_cache(maxsize=maxsize)(func)

        @wraps(func)
        def inner(call: Any):
            if not callable(call):
                return False
            if inspect.isclass(call):
                return False
            if isinstance(call, functools.partial):
                call = call.func
            if func(call):
                return True
            _call = getattr(call, "__call__", None)
            if _call is None:
                return False
            return func(_call)

        return inner  # type: ignore[return-value]

    return wrapper


@cached_accept_callable_class(maxsize=2048)
def is_coroutine_callable(call: DependencyProvider) -> bool:
    return inspect.iscoroutinefunction(call)


@cached_accept_callable_class(maxsize=2048)
def is_async_gen_callable(call: DependencyProvider) -> bool:
    return inspect.isasyncgenfunction(call)


@cached_accept_callable_class(maxsize=2048)
def is_gen_callable(call: Any) -> bool:
    return inspect.isgeneratorfunction(call)


@lru_cache(maxsize=2048)
def get_annotations(call: DependencyProvider) -> Dict[str, Any]:
    types_from: DependencyProvider
    if inspect.isclass(call):
        types_from = call.__init__  # type: ignore[misc] # accessing __init__ directly
    elif not (inspect.isfunction(call) or inspect.ismethod(call)) and hasattr(
        call, "__call__"
    ):
        # callable class
        types_from = call.__call__  # type: ignore[misc,operator] # accessing __init__ directly
    else:
        # method
        types_from = call
    return get_type_hints(types_from)


@lru_cache(maxsize=2048)
def get_parameters(call: DependencyProvider) -> Dict[str, inspect.Parameter]:
    params: Mapping[str, inspect.Parameter]
    if inspect.isclass(call) and call.__new__ is not object.__new__:
        # classes overriding __new__, including some generic metaclasses, result in __new__ getting read
        # instead of __init__
        params = inspect.signature(call.__init__).parameters  # type: ignore[misc] # accessing __init__ directly
        params = dict(params)
        params.pop(next(iter(params.keys())))  # first parameter to __init__ is self
    else:
        params = inspect.signature(call).parameters
    annotations: Optional[Dict[str, Any]] = None
    processed_params: Dict[str, inspect.Parameter] = {}
    for param_name, param in params.items():
        if isinstance(param.annotation, str):
            if annotations is None:
                annotations = get_annotations(call)
            param = param.replace(
                annotation=annotations.get(param_name, param.annotation)
            )
        processed_params[param_name] = param
    return processed_params


def infer_call_from_annotation(parameter: inspect.Parameter) -> DependencyProvider:
    if not callable(parameter.annotation):
        raise WiringError(
            f"Annotation for {parameter.name} is not a callable class or function and so we cannot autowire it."
            " You must explicity provide a default value implementing the DependencyProtocol"
        )
    return parameter.annotation
